#version 430 core

// Input variables
layout (location = 0) in vec4 position;
layout (location = 1) in vec3 normal;

// Output variables
out VertexData
{
	vec3 normal_ws;		// Normal in world space
	vec3 position_ws;	// Position in world space
	vec3 color;
} outData;

// Data of the camera
layout (std140, binding = 0) uniform CameraData
{
	mat4 projection;		// Projection matrix
	mat4 projection_inv;	// Inverse of the projection matrix
	mat4 view;				// View matrix
	mat4 view_inv;			// Inverse of the view matrix
	mat3 view_it;			// Inverse of the transpose of the top-left part 3x3 of the view matrix
	vec3 eye_position;		// Position of the eye in world space
};

// Data of the objects
struct SingleModelData
{
	mat4 model;			// Model matrix
	mat4 model_inv;		// Inverse of the model matrix
	mat3 model_it;		// Inverse of the transpose of the top-left part 3x3 of the model matrix
};
layout (std140, binding = 2) uniform ModelData
{
	SingleModelData models[9];
};

// Data of the materials
struct PhongMaterial
{
	// See the C++ code for the documentation to individual attributes
	vec3 ambient;
	vec3 diffuse;
	float alpha;
	vec3 specular;
	float shininess;
};
layout (std140, binding = 3) uniform MaterialData
{
	PhongMaterial materials[9];
};

//-----------------------------------------------------------------------

void main()
{
	outData.position_ws = vec3(models[gl_InstanceID].model * position);
	outData.normal_ws = normalize(models[gl_InstanceID].model_it * normal);
	
	// I pass only the diffuse color. Ambient is the same as diffuse, and other parameters are the same for all materials
	outData.color = materials[gl_InstanceID].diffuse;

	gl_Position = projection * view * models[gl_InstanceID].model * position;
}
