#include "repetition_main.h"

//---------------------
//----    SCENE    ----
//---------------------

/// Resizes the textures of the framebuffer object according to the size of the window
void resize_fbo_textures()
{
	fbo_width = win_width;
	fbo_height = win_height;

	glBindTexture(GL_TEXTURE_2D, fbo_color_texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, fbo_width, fbo_height, 0, GL_RGBA, GL_UNSIGNED_BYTE, nullptr);
	glBindTexture(GL_TEXTURE_2D, fbo_depth_stencil_texture);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_DEPTH24_STENCIL8, fbo_width, fbo_height, 0, GL_DEPTH_STENCIL, GL_UNSIGNED_INT_24_8, nullptr);
	glBindTexture(GL_TEXTURE_2D, 0);
}

/// Reloads all shaders
void reload_shaders()
{
	// This code contains a lot of commented code. The code would be necessary if we did not use
	// layout (location = x) and layout (binding = x) in shaders.

	// Cube program
	cube_program.Init();
	cube_program.AddVertexShader("Shaders/cube_vertex.glsl");
	cube_program.AddFragmentShader("Shaders/cube_fragment.glsl");
	//cube_program.BindAttribLocation(DEFAULT_POSITION_LOC, "position");
	//cube_program.BindAttribLocation(DEFAULT_NORMAL_LOC, "normal");
	//cube_program.BindAttribLocation(DEFAULT_TEX_COORD_LOC, "tex_coord");
	cube_program.Link();

	//cube_program.SetUniformBlockBindingIndex("CameraData", DEFAULT_CAMERA_BINDING);
	//cube_program.SetUniformBlockBindingIndex("PhongLightsData", DEFAULT_LIGHTS_BINDING);
	//cube_program.SetUniformBlockBindingIndex("CubeData", DEFAULT_OBJECT_BINDING);
	//cube_program.SetUniformBlockBindingIndex("MaterialData", DEFAULT_MATERIAL_BINDING);

	//if (cube_program.IsValid())
	//{
	//	cube_program.Use();
	//	cube_program.Uniform1i("wood_tex", 0);
	//	glUseProgram(0);
	//}

	// Teapot program
	teapot_program.Init();
	teapot_program.AddVertexShader("Shaders/teapot_vertex.glsl");
	teapot_program.AddFragmentShader("Shaders/teapot_fragment.glsl");
	//teapot_program.BindAttribLocation(DEFAULT_POSITION_LOC, "position");
	//teapot_program.BindAttribLocation(DEFAULT_NORMAL_LOC, "normal");
	teapot_program.Link();

	//teapot_program.SetUniformBlockBindingIndex("CameraData", DEFAULT_CAMERA_BINDING);
	//teapot_program.SetUniformBlockBindingIndex("PhongLightsData", DEFAULT_LIGHTS_BINDING);
	//teapot_program.SetUniformBlockBindingIndex("TeapotData", DEFAULT_OBJECT_BINDING);
	//teapot_program.SetUniformBlockBindingIndex("MaterialData", DEFAULT_MATERIAL_BINDING);

	// Program for objects without textures and without instancing
	notexture_program.Init();
	notexture_program.AddVertexShader("Shaders/notexture_vertex.glsl");
	notexture_program.AddFragmentShader("Shaders/notexture_fragment.glsl");
	//notexture_program.BindAttribLocation(DEFAULT_POSITION_LOC, "position");
	//notexture_program.BindAttribLocation(DEFAULT_NORMAL_LOC, "normal");
	notexture_program.Link();

	//notexture_program.SetUniformBlockBindingIndex("CameraData", DEFAULT_CAMERA_BINDING);
	//notexture_program.SetUniformBlockBindingIndex("PhongLightsData", DEFAULT_LIGHTS_BINDING);
	//notexture_program.SetUniformBlockBindingIndex("ModelData", DEFAULT_OBJECT_BINDING);
	//notexture_program.SetUniformBlockBindingIndex("MaterialData", DEFAULT_MATERIAL_BINDING);

	// Postprocessing program
	postprocessing_program.Init();
	postprocessing_program.AddVertexShader("Shaders/postprocessing_vertex.glsl");
	postprocessing_program.AddFragmentShader("Shaders/postprocessing_fragment.glsl");
	postprocessing_program.Link();

	//if (postprocessing_program.IsValid())
	//{
	//	postprocessing_program.Use();
	//	postprocessing_program.Uniform1i("input_tex", 0);
	//	glUseProgram(0);
	//}

	cout << "Shaders are reloaded" << endl;
}

/// Initializes all objects of our scene
void init_scene()
{
	//----------------------------------------------
	//--  Display some information about the OpenGL

	PrintOpenGLInfo();

	//----------------------------------------------
	//--  Load shaders

	reload_shaders();
	
	//----------------------------------------------
	//--  Prepare the lights

	PhongLights_ubo.Init();
	PhongLights_ubo.SetGlobalAmbient(glm::vec3(0.0f));

	//----------------------------------------------
	//--  Prepare materials

	CubeMaterial_ubo.Init();
	CubeMaterial_ubo.SetMaterial(PhongMaterial::CreateBasicMaterial(glm::vec3(1.0f), true, 40.0f, 1.0f));
	CubeMaterial_ubo.UpdateOpenGLData();

	FloorMaterial_ubo.Init();
	FloorMaterial_ubo.SetMaterial(PhongMaterial::CreateBasicMaterial(glm::vec3(0.7f), false, 40.0f, 1.0f));
	FloorMaterial_ubo.UpdateOpenGLData();

	GreenMaterial_ubo.Init();
	GreenMaterial_ubo.SetMaterial(PhongMaterial::CreateBasicMaterial(glm::vec3(0.0f, 1.0f, 0.0f), true, 40.0f, 1.0f));
	GreenMaterial_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare all cameras

	the_camera.SetCamera(-0.4f, -0.7f, 30.0f);
	CameraData_ubo.Init();

	//----------------------------------------------
	//--  Positions of objects in the scene

	FloorModel_ubo.Init();
	FloorModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, -1.1f, 0.0f)) *
		glm::scale(glm::mat4(1.0f), glm::vec3(15.0f, 0.1f, 15.0f)));
	FloorModel_ubo.UpdateOpenGLData();

	SphereModel_ubo.Init();
	SphereModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), glm::vec3(-9.0f, 0.0f, 9.0f)));
	SphereModel_ubo.UpdateOpenGLData();

	TorusModel_ubo.Init();
	TorusModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), glm::vec3(9.0f, -0.5f, 9.0f)));
	TorusModel_ubo.UpdateOpenGLData();

	CylinderModel_ubo.Init();
	CylinderModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), glm::vec3(-9.0f, -0.5f, -9.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f)));
	CylinderModel_ubo.UpdateOpenGLData();

	CapsuleModel_ubo.Init();
	CapsuleModel_ubo.SetMatrix(
		glm::translate(glm::mat4(1.0f), glm::vec3(9.0f, -0.5f, -9.0f)) *
		glm::rotate(glm::mat4(1.0f), glm::radians(90.0f), glm::vec3(1.0f, 0.0f, 0.0f)));
	CapsuleModel_ubo.UpdateOpenGLData();

	// Prepare instancing data of cubes and teapots:
	CubesModel_ubo.Init(INSTANCES_COUNT);
	TeapotsModel_ubo.Init(INSTANCES_COUNT);
	TeapotsMaterial_ubo.Init(INSTANCES_COUNT);
	
	// Compute the data of all instances
	srand(12345);		// Any seed we want
	for (int i = 0; i < INSTANCES_COUNT; i++)
	{
		// Compute the positions of all cubes - create a 3x3 grid
		glm::mat4 cube_mat = 
			glm::translate(glm::mat4(1.0f), glm::vec3(float(i%3 - 1) * 5.0f, 0.0f, float(i/3 - 1) * 5.0f));

		// Compute the position of all teapots - place them on the cubes and randomly rotate
		glm::mat4 teapot_mat = 
			cube_mat *
			glm::translate(glm::mat4(1.0f), glm::vec3(0.0f, 1.0f, 0.0f)) *
			glm::rotate(glm::mat4(1.0f), float(rand()) / float(RAND_MAX) * 2.0f * float(M_PI), glm::vec3(0.0f, 1.0f, 0.0f));

		// Choose a random color for each teapot
		glm::vec3 teapot_color = glm::vec3(float(rand()), float(rand()), float(rand())) / float(RAND_MAX);

		// Set the data
		CubesModel_ubo.SetMatrix(i, cube_mat);
		TeapotsModel_ubo.SetMatrix(i, teapot_mat);
		TeapotsMaterial_ubo.SetMaterial(i, PhongMaterial::CreateBasicMaterial(teapot_color, true, 40.0f, 1.0f));
	}

	CubesModel_ubo.UpdateOpenGLData();
	TeapotsModel_ubo.UpdateOpenGLData();
	TeapotsMaterial_ubo.UpdateOpenGLData();

	//----------------------------------------------
	//--  Prepare geometries

	geom_cube = CreateCube();
	geom_sphere = CreateSphere();
	geom_torus = CreateTorus();
	geom_cylinder = CreateCylinder();
	geom_capsule = CreateCapsule();
	geom_teapot = CreateTeapot();
	geom_fullscreen_quad = CreateFullscreenQuad();

	//----------------------------------------------
	//--  Prepare textures

	// Load and set the parameters of the wood texture
	wood_tex = CreateAndLoadTexture2D(L"../../textures/wood.jpg");
	glBindTexture(GL_TEXTURE_2D, wood_tex);
	SetTexture2DParameters(GL_REPEAT, GL_REPEAT, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	//----------------------------------------------
	//--  Prepare framebuffers

	// Prepare FBO textures
	glGenTextures(1, &fbo_color_texture);
	glGenTextures(1, &fbo_depth_stencil_texture);
	resize_fbo_textures();
	
	// Set parameters of the color texture. We do not use the depth texture for rendering,
	// so we do not set its parameters.
	glBindTexture(GL_TEXTURE_2D, fbo_color_texture);
	SetTexture2DParameters(GL_CLAMP_TO_EDGE, GL_CLAMP_TO_EDGE, GL_LINEAR, GL_LINEAR);
	glBindTexture(GL_TEXTURE_2D, 0);

	// Prepare FBO
	glGenFramebuffers(1, &fbo);
	glBindFramebuffer(GL_FRAMEBUFFER, fbo);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, fbo_color_texture, 0);
	glFramebufferTexture2D(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_TEXTURE_2D, fbo_depth_stencil_texture, 0);
	glDrawBuffers(1, DrawBuffersConstants);
	CheckFramebufferStatus("Postprocessing");
	glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

/// Updates the scene: performs animations, updates the data of the buffers, etc.
void update_scene(int app_time_diff_ms)
{
	// Data of the lights
	PhongLights_ubo.PhongLights.clear();
	glm::vec3 atten_factors = use_attenuation ? glm::vec3(atten_constant, atten_linear, atten_quadratic) : glm::vec3(1.0f, 0.0f, 0.0f);
	if (use_point_light)
	{
		PhongLights_ubo.PhongLights.push_back(PhongLight::CreatePointLight(
			glm::vec3(0.0f, 5.0f, 5.0f),
			glm::vec3(0.04f, 0.04f, 0.04f),
			glm::vec3(0.7f, 0.7f, 0.7f),
			glm::vec3(0.7f, 0.7f, 0.7f),
			atten_factors.x, atten_factors.y, atten_factors.z));
	}
	if (use_directional_light)
	{
		PhongLights_ubo.PhongLights.push_back(PhongLight::CreateDirectionalLight(
			glm::vec3(1.0f, 1.0f, 0.0f),
			glm::vec3(0.04f, 0.04f, 0.04f),
			glm::vec3(0.4f, 0.4f, 0.4f),
			glm::vec3(0.4f, 0.4f, 0.4f)));
	}
	if (use_spot_lights)
	{
		PhongLights_ubo.PhongLights.push_back(PhongLight::CreateSpotLight(
			glm::vec3(-5.0f, 5.0f, 5.0f),
			glm::vec3(0.04f, 0.04f, 0.04f),
			glm::vec3(0.7f, 0.7f, 0.7f),
			glm::vec3(0.7f, 0.7f, 0.7f),
			-glm::normalize(glm::vec3(-5.0f, 5.0f, 5.0f)), spot_exponent, cosf(glm::radians(spot_cutoff_deg)),
			atten_factors.x, atten_factors.y, atten_factors.z));
		PhongLights_ubo.PhongLights.push_back(PhongLight::CreateSpotLight(
			glm::vec3(5.0f, 5.0f, 5.0f),
			glm::vec3(0.04f, 0.04f, 0.04f),
			glm::vec3(0.7f, 0.7f, 0.7f),
			glm::vec3(0.7f, 0.7f, 0.7f),
			-glm::normalize(glm::vec3(5.0f, 5.0f, 5.0f)), spot_exponent, cosf(glm::radians(spot_cutoff_deg)),
			atten_factors.x, atten_factors.y, atten_factors.z));
	}
	PhongLights_ubo.UpdateOpenGLData();

	// Data of the camera
	CameraData_ubo.SetProjection(
		glm::perspective(glm::radians(45.0f), float(win_width) / float(win_height), 0.1f, 1000.0f));
	CameraData_ubo.SetCamera(the_camera);
	CameraData_ubo.UpdateOpenGLData();
}

/// Renders the whole frame
void render_scene()
{
	if (postprocessing)
		glBindFramebuffer(GL_FRAMEBUFFER, fbo);		// When postprocessing, draw into FBO
	else
		glBindFramebuffer(GL_FRAMEBUFFER, 0);		// When not postprocessing, draw directly into the window

	// Set the clear values and clear the framebuffer
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearDepth(1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glEnable(GL_DEPTH_TEST);

	// Set the data of the scene, common for all objects - the camera and the lights
	CameraData_ubo.BindBuffer(DEFAULT_CAMERA_BINDING);
	PhongLights_ubo.BindBuffer(DEFAULT_LIGHTS_BINDING);

	if (wireframe)
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

	// Render the cubes:
	if (cube_program.IsValid())
	{
		//	- set the program
		cube_program.Use();

		//	- bind the buffer with the data of the cubes
		CubesModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);

		//	- bind the buffer with the basic material of the cubes
		CubeMaterial_ubo.BindBuffer(DEFAULT_MATERIAL_BINDING);

		//	- bind the textures
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, wood_tex);

		//	- draw the cubes
		geom_cube.BindVAO();
		geom_cube.DrawInstanced(INSTANCES_COUNT);
	}

	// Render the teapots:
	if (teapot_program.IsValid())
	{
		//	- set the program
		teapot_program.Use();

		//	- bind the buffer with the data of the teapots
		TeapotsModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);

		//	- bind the buffer with the basic material of the teapots
		TeapotsMaterial_ubo.BindBuffer(DEFAULT_MATERIAL_BINDING);

		//	- draw the teapot
		geom_teapot.BindVAO();
		geom_teapot.DrawInstanced(INSTANCES_COUNT);
	}

	// Render the floor and other objects
	if (notexture_program.IsValid())
	{
		//	- set the program
		notexture_program.Use();

		//	- bind the buffer with the data of the floor
		FloorModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);
		//	- bind the buffer with the material of the floor
		FloorMaterial_ubo.BindBuffer(DEFAULT_MATERIAL_BINDING);

		//	- draw the floor
		geom_cube.BindVAO();
		geom_cube.Draw();

		// Draw other objects: sphere, torus, cylinder, and capsule. They have their own geometries
		// and positions, but they share the same shaders and the same material.

		GreenMaterial_ubo.BindBuffer(DEFAULT_MATERIAL_BINDING);

		SphereModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);
		geom_sphere.BindVAO();
		geom_sphere.Draw();

		TorusModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);
		geom_torus.BindVAO();
		geom_torus.Draw();

		CylinderModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);
		geom_cylinder.BindVAO();
		geom_cylinder.Draw();

		CapsuleModel_ubo.BindBuffer(DEFAULT_OBJECT_BINDING);
		geom_capsule.BindVAO();
		geom_capsule.Draw();
	}

	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// Postprocessing:
	if (postprocessing)
	{
		//	- draw into the main window
		glBindFramebuffer(GL_FRAMEBUFFER, 0);

		//	- set the clear values and clear the framebuffer. We do not need the depth buffer.
		glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		glClear(GL_COLOR_BUFFER_BIT);
		
		//	- we do not need the depth test
		glDisable(GL_DEPTH_TEST);
		
		//	- bind the texture with the scene we just rendered
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, fbo_color_texture);

		//	- bind the program
		postprocessing_program.Use();

		//	- bind the empty VAO (the geometry is defined in vertex shader) and render
		//		a triangle spanning the whole screen
		geom_fullscreen_quad.BindVAO();
		geom_fullscreen_quad.Draw();
	}

	// Reset the VAO and the program
	glBindVertexArray(0);
	glUseProgram(0);
}

//-------------------
//----    GUI    ----
//-------------------

/// Callback when user clicks on Reload button
void TW_CALL reload(void *)
{
	reload_shaders();
}

void init_gui()
{
	// Initialize AntTweakBar
	TwInit(TW_OPENGL_CORE, nullptr);
	TwWindowSize(win_width, win_height);
	TwGLUTModifiersFunc(glutGetModifiers);

	// Initialize GUI
	the_gui = TwNewBar("Parameters");
	TwAddButton(the_gui, "Reload", reload, nullptr, nullptr);
	TwAddVarRW(the_gui, "Wireframe", TW_TYPE_BOOLCPP, &wireframe, nullptr);
	TwAddVarRW(the_gui, "Postprocessing", TW_TYPE_BOOLCPP, &postprocessing, nullptr);
	TwAddVarRW(the_gui, "Use point light", TW_TYPE_BOOLCPP, &use_point_light, nullptr);
	TwAddVarRW(the_gui, "Use directional light", TW_TYPE_BOOLCPP, &use_directional_light, nullptr);
	TwAddVarRW(the_gui, "Use spot lights", TW_TYPE_BOOLCPP, &use_spot_lights, nullptr);
	TwAddVarRW(the_gui, "Use light attentuation", TW_TYPE_BOOLCPP, &use_attenuation, nullptr);
	TwAddVarRW(the_gui, "Spot Cutoff (degress)", TW_TYPE_FLOAT, &spot_cutoff_deg, "min=5 max=80 step=1");
	TwAddVarRW(the_gui, "Spot Exponent ", TW_TYPE_FLOAT, &spot_exponent, "min=0 max=80 step=0.2");
	TwAddVarRW(the_gui, "Constant Attenuation", TW_TYPE_FLOAT, &atten_constant, "min=0 max=1 step=0.01");
	TwAddVarRW(the_gui, "Linear Attenuation", TW_TYPE_FLOAT, &atten_linear, "min=0 max=1 step=0.01");
	TwAddVarRW(the_gui, "Quadratic Attenuation", TW_TYPE_FLOAT, &atten_quadratic, "min=0 max=1 step=0.001");
}

//---------------------------
//----    APPLICATION    ----
//---------------------------

/// GLUT callback - when the user presses a key. See glutKeyboardFunc for more info
void on_keyboard_func(unsigned char key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventKeyboardGLUT(key, x, y))
		return;		// Already handled by AntTweakBar

	switch (key)
	{
	case 27:		// Escape key
		exit(0);
		break;
	}
}

/// GLUT callback - when the user presses a special key (like F1 key). See glutSpecialFunc for more info
void on_special_func(int key, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventSpecialGLUT(key, x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the user presses or releases a mouse button. See glutMouseFunc for more info
void on_mouse_func(int button, int state, int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseButtonGLUT(button, state, x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMouseFunc(button, state, x, y);
}

/// GLUT callback - when the user moves the mouse when holding a botton. See glutMotionFunc for more info
void on_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar

	the_camera.OnMotionFunc(x, y);
}

/// GLUT callback - when the user moves the mouse when not holding a botton. See glutPassiveMotionFunc for more info
void on_passive_motion_func(int x, int y)
{
	// Inform AntTweakBar
	if (TwEventMouseMotionGLUT(x, y))
		return;		// Already handled by AntTweakBar
}

/// GLUT callback - when the window is resized. See glutReshapeFunc for more info
void on_reshape(int width, int height)
{
	win_width = width;
	win_height = height;
	
	glViewport(0, 0, win_width, win_height);
	
	// Inform AntTweakBar
	TwWindowSize(win_width, win_height);

	// Resize FBOs
	resize_fbo_textures();
}

/// GLUT callback - when our animation timer requests a new frame. See glutTimerFunc for more info
void on_timer(int)
{
	// Register the timer to be called again in 20 ms, and rerender the scene
	glutTimerFunc(20, on_timer, 0);
	glutPostRedisplay();
}

/// GLUT callback - when the window needs to be redrawn. See glutDisplayFunc for more info
void on_display()
{
	//--  Update all the data

	// Update the application time
	int current_glut_time = glutGet(GLUT_ELAPSED_TIME);
	int app_time_diff_ms = current_glut_time - last_glut_time;
	app_time_ms += app_time_diff_ms;
	last_glut_time = current_glut_time;

	// Update the scene
	update_scene(app_time_diff_ms);

	//--  Render the scene

	// Render the scene
	render_scene();
	
	// Render the GUI
	TwDraw();

	// Swaps the front and back buffer (double-buffering)
	glutSwapBuffers();
}

/// Callback function to be called when we make an error in OpenGL
void GLAPIENTRY simple_debug_callback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const char* message, const void* userParam)
{
	switch (type)			// Report only some errors
	{
	case GL_DEBUG_TYPE_ERROR:
	case GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR:
	case GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR:
		cout << message << endl;		// Put the breakpoint here
		return;
	default:
		return;
	}
}

/// C main function :-)
int main(int argc, char ** argv)
{
	// Initialize GLUT
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGBA);
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);		// Return from the main loop

	// Set OpenGL Context parameters
	glutInitContextVersion(4, 3);				// Use OpenGL 4.3
	glutInitContextProfile(GLUT_CORE_PROFILE);	// Use OpenGL core profile
	glutInitContextFlags(GLUT_DEBUG);			// Use OpenGL debug context

	// Initialize window
	glutInitWindowSize(win_width, win_height);
	glutCreateWindow("PV227 - Repetition of PV112 knowledge");

	// Initialize GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	// Initialize DevIL library
	InitDevIL();

	// Set the OpenGL debug callback
	SetDebugCallback(simple_debug_callback);

	// Register GLUT callbacks
	glutDisplayFunc(on_display);
	glutReshapeFunc(on_reshape);
	glutTimerFunc(20, on_timer, 0);

	glutKeyboardFunc(on_keyboard_func);
	glutSpecialFunc(on_special_func);
	glutMouseFunc(on_mouse_func);
	glutMotionFunc(on_motion_func);
	glutPassiveMotionFunc(on_passive_motion_func);

	// Initialize OpenGL stuff
	init_gui();
	init_scene();

	// Everything is loaded and prepared, reset the application time
	last_glut_time = glutGet(GLUT_ELAPSED_TIME);
	app_time_ms = 0;

	// Run the main loop
	glutMainLoop();

	// Unload AntTweakBar
	TwTerminate();

	return 0;
}