// PV227, Repetition of PV112 knowledge
#pragma once

#include "../../Framework/PV227.h"

using namespace std;
using namespace PV227;

//---------------------
//----    SCENE    ----
//---------------------

// Number of instances of cubes and teapots
static const int INSTANCES_COUNT = 9;

// Shader programs for individual objects
ShaderProgram cube_program;
ShaderProgram teapot_program;
ShaderProgram notexture_program;

// Shader program for post-processing
ShaderProgram postprocessing_program;

// Geometries we use in this lecture
Geometry geom_cube;
Geometry geom_sphere;
Geometry geom_torus;
Geometry geom_cylinder;
Geometry geom_capsule;
Geometry geom_teapot;
Geometry geom_fullscreen_quad;

// A camera that allows us to look at the objects from different views
SimpleCamera the_camera;

// Wood texture
GLuint wood_tex;

// Data of our lights - positions, colors, etc.
PhongLightsData_UBO PhongLights_ubo;

// Data of our camera - view matrix, projection matrix, etc.
CameraData_UBO CameraData_ubo;

// Data of our materials
MaterialData_UBO CubeMaterial_ubo;
MaterialData_UBO FloorMaterial_ubo;
MaterialData_UBO GreenMaterial_ubo;
MaterialData_UBO TeapotsMaterial_ubo;

// Data of our objects (their positions)
ModelData_UBO FloorModel_ubo;
ModelData_UBO SphereModel_ubo;
ModelData_UBO TorusModel_ubo;
ModelData_UBO CylinderModel_ubo;
ModelData_UBO CapsuleModel_ubo;
ModelData_UBO CubesModel_ubo;
ModelData_UBO TeapotsModel_ubo;

// Framebuffer object for postprocessing
int fbo_width, fbo_height;
GLuint fbo_color_texture;
GLuint fbo_depth_stencil_texture;
GLuint fbo;

// Functions that works with scene objects
void resize_fbo_textures();
void reload_shaders();
void init_scene();
void update_scene(int app_time_diff_ms);
void render_scene();

//-------------------
//----    GUI    ----
//-------------------

// Main AntTweakBar object
TwBar *the_gui;

// Variables that are changed with GUI
bool wireframe = false;
bool postprocessing = false;
bool use_point_light = true;
bool use_directional_light = true;
bool use_spot_lights = true;
bool use_attenuation = false;
float spot_cutoff_deg = 20.0f;
float spot_exponent = 20.0f;
float atten_constant = 0.1f;
float atten_linear = 0.1f;
float atten_quadratic = 0.01f;

// Callbacks from the GUI
void TW_CALL reload(void *);

// Functions that works with GUI
void init_gui();

//---------------------------
//----    APPLICATION    ----

// Current window size
int win_width = 640;
int win_height = 480;

// Current time of the application in milliseconds, for animations
int app_time_ms = 0;
int last_glut_time = 0;
